""" :mod:`hedwig.backends.base`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
from threading import RLock


class BaseEmailBackend(object):
    """
    Base class ofr email backend implementation.
    """

    def __init__(self, **kwargs):
        self.opened = 0
        self._lock = RLock()

    def open(self):
        self.opened += 1

    def close(self):
        if self.opened:
            self.opened -= 1
        else:
            raise RuntimeError('already closed')

    def send_messages(self, email_messages):
        raise NotImplementedError()

    def __enter__(self):
        if not self.opened:
            self._lock.acquire()
            self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        self._lock.release()

    def __repr__(self):
        cls = type(self)
        mod = cls.__module__
        mod = '' if __name__ == mod else mod + '.'
        kwargs = ((k, getattr(self, k)) for k in dir(self)
                                        if not k.startswith('_'))
        kwargs = [(k, v) for k, v in kwargs if not callable(v)]
        kwargs.sort(key=lambda (k, v): k)
        kwargs = ', '.join('{0}={1!r}'.format(k, v) for k, v in kwargs)
        return '{0}{1}({2})'.format(mod, cls.__name__, kwargs)



""" :mod:`hedwig.backends.console`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
from sys import stderr, stdout
from threading import RLock

from .base import BaseEmailBackend


class ConsoleEmailBackend(BaseEmailBackend):
    """
    Console I/O email backend implementation.
    """

    def __init__(self, stream=None, **kwargs):
        super(ConsoleEmailBackend, self).__init__(**kwargs)
        self.stream = stream or stderr
        self._lock = RLock()

    def send_messages(self, email_messages):
        if not email_messages:
            return
        for message in email_messages:
            self.stream.write('%s\n' % message.message().as_string())
            self.stream.write('-' * 79)
            self.stream.write('\n')
            self.stream.flush()
        return len(email_messages)

    def __repr__(self):
        cls = type(self)
        mod = cls.__module__
        mod = '' if __name__ == mod else mod + '.'
        if self.stream == stdout:
            stream = 'sys.stdout'
        elif self.stream == stderr:
            stream = 'sys.stderr'
        else:
            stream = repr(stream)
        kwargs = [(k, getattr(self, k)) for k in dir(self)]
        kwargs = [(k, v) for k, v in kwargs if not callable(v)]
        kwargs.append(('stream', stream))
        kwargs.sort(key=lambda (k, v): k)
        kwargs = ', '.join('{0}={1!r}'.format(k, v) for k, v in kwargs)
        return '{0}{1}({2})'.format(mod, cls.__name__, kwargs)



""" :mod:`hedwig.message`
    ~~~~~~~~~~~~~~~~~~~~~

"""
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Utils import formatdate
from os import getpid
from random import randrange
from socket import getfqdn
from sys import getdefaultencoding
from time import gmtime, strftime, time


SUPPORTED_CONTENT_SUBTYPES = frozenset(['plain', 'html'])


def make_message_id(idstring=None):
    timeval = time()
    utcdate = strftime('%Y%m%d%H%M%S', gmtime(timeval))
    pid = getpid()
    randint = randrange(100000)
    if idstring is None:
        idstring = ''
    else:
        idstring = '.' + idstring
    DNS_NAME = getfqdn()
    msgid = '<{0}.{1}.{2}{3}@{4}>'.format(utcdate, pid, randint, idstring,
                                          DNS_NAME)


class EmailMessage(object):
    """
    Email.
    """

    def __init__(self, subject='', body='', from_=None, to=None, bcc=None,
                 attachments=(), headers={}, encoding=None,
                 content_subtype='html'):
        if from_ and not isinstance(from_, basestring):
            raise TypeError('from_ must be a string, not ' + repr(from_))
        if to:
            if not isinstance(to, (list, tuple)):
                to = [to]
            for each in to:
                if not isinstance(each, basestring):
                    raise TypeError('to must be a list or tuple contains '
                                    'basestring')
        self.to = to or []
        if bcc:
            if not isinstance(bcc, (list, tuple)):
                bcc = [bcc]
            for each in bcc:
                if not isinstance(each, basestring):
                    raise TypeError('bcc must be a list or tuple contains '
                                    'basestring')
        self.bcc = bcc or []
        self.from_ = from_
        self.subject = subject
        if encoding is None and isinstance(body, unicode):
            body = body.encode('utf-8')
            encoding = 'utf-8'
        self.body = body
        self.attachments = []
        for file_ in attachments:
            raise NotImplementedError()
        self.extra_headers = dict(headers)
        self.encoding = encoding or getdefaultencoding()
        if content_subtype not in SUPPORTED_CONTENT_SUBTYPES:
            supports = SUPPORTED_CONTENT_SUBTYPES
            if len(supports) == 1:
                supports = supports[0]
            else:
                supports = ', '.join(supports[:-1]) + ' or %s' % supports[-1]
            raise ValueError('content_subtype must be %s' % supports)
        self.content_subtype = content_subtype

    def message(self):
        """ A message of the email. """
        body = self.body
        if not isinstance(body, str):
            body = unicode(body) if hasattr(body, '__unicode__') else str(body)
        if isinstance(body, unicode):
            body = body.encode(self.encoding)
        msg = MIMEText(body, self.content_subtype, self.encoding)
        msg = self._create_message(msg)
        msg['Subject'] = self.subject
        msg['From'] = self.extra_headers.get('From', self.from_)
        msg['To'] = ', '.join(self.to)
        header_names = [key.lower() for key in self.extra_headers]
        if 'date' not in header_names:
            msg['Date'] = formatdate()
        if 'message-id' not in header_names:
            msg['Message-ID'] = make_message_id()
        for name, value in self.extra_headers.items():
            if name.lower() == 'from':
                continue
            msg[name] = value
        return msg

    def _create_message(self, message):
        if self.attachments:
            raise NotImplementedError()
        return message

    @property
    def recipients(self):
        """ A list of all recipients of the email. """
        return self.to + self.bcc

    def attach(self, *args):
        raise NotImplementedError()

    def __repr__(self):
        cls = type(self)
        mod = cls.__module__
        mod = '' if __name__ == mod else mod + '.'
        args = [
            ('subject', self.subject),
            ('body', self.body),
            ('from_', self.from_),
            ('to', self.to),
            ('bcc', self.bcc),
            ('attachments', self.attachments),
            ('headers', self.extra_headers),
            ('encoding', self.encoding),
        ]
        kwargs = ', '.join('{0}={1!r}'.format(k, v) for k, v in args)
        return '{0}{1}({2})'.format(mod, cls.__name__, kwargs)



""" :mod:`hedwig.backends.smtp`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
from smtplib import SMTP
from socket import getfqdn

from .base import BaseEmailBackend


class SMTPEmailBackend(BaseEmailBackend):
    """
    SMTP email backend implementation.
    """

    def __init__(self, host='localhost', port=25, username=None, password=None,
                 use_tls=False, **kwargs):
        super(SMTPEmailBackend, self).__init__(**kwargs)
        if isinstance(port, basestring):
            try: 
                port = int(port)
            except ValueError:
                pass
        if not isinstance(host, basestring):
            raise TypeError('host must be a string, not ' + repr(host))
        if not isinstance(port, int):
            raise TypeError('port must be an integer, not ' + repr(port))
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.use_tls = use_tls
        self.connection = None

    def open(self):
        super(SMTPEmailBackend, self).open()
        for trial in xrange(5):
            self._open_connection()
            if self.connection:
                break
        if not self.connection:
            raise RuntimeError('cannot open smtp connection') 

    def _open_connection(self):
        local_hostname = getfqdn()
        self.connection = SMTP(self.host, self.port,
                               local_hostname=local_hostname)
        if self.use_tls:
            self.connection.ehlo()
            self.connection.starttls()
            self.connection.ehlo()
        if self.username and self.password:
            self.connection.login(self.username, self.password)

    def close(self):
        if not isinstance(self.connection, SMTP):
            raise RuntimeError('already closed or not opened ever')
        try:
            self.connection.quit()
        except socket.sslerror:
            self.connection.close()
        finally:
            self.connection = None
        super(SMTPEmailBackend, self).close()

    def send_messages(self, email_messages):
        if not email_messages:
            return
        n_sent = 0
        for message in email_messages:
            sent = self._send_message(message)
            if sent:
                n_sent += 1
        return n_sent

    def _send_message(self, email_message):
        if not email_message.recipients:
            return False
        self.connection.sendmail(email_message.from_,
                                 email_message.recipients,
                                 email_message.message().as_string())
        return True



from setuptools import find_packages, setup


with open('README.rst') as f:
    long_description = f.read()


setup(
    name='Hedwig',
    version='0.1',
    url='https://github.com/kimjayd/hedwig',
    license='MIT',
    author='Hyunjun Kim',
    author_email='kim@hyunjun.kr',
    description='django.core.mail without django.',
    long_description=long_description,
    packages=find_packages(),
    platforms='any',
)
